% jbn's nn mode.                            % -*- mode: slang; mode: fold; -*-
% 
% Copyright 2000-2007 J.B. Nicholson-Owens
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program; if not, write to the Free Software
%    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%
%    You can contact the author at <jbn@forestfield.org>.  A copy of the
%    applicable license is also available on the author's website
%    http://forestfield.org/ free of charge.
%
% Author: J.B. Nicholson-Owens
% Name: nn.sl
% License: GNU GPL
% Version: 0.1 (pre-alpha)
%
%{{{
% This is another nn mode for slrn.  It doesn't support everything the real
% nn does but it should be helpful to those nn users just getting started
% in slrn.  This is not intended to be a replacement or emulation of nn.  It
% is intended to assist those switching from nn to slrn by using familiar
% nn concepts to introduce slrn's capabilities.
%
% The most significant differences between nn and this nn front-end are:
% 
% * There is still much to be implemented in this front-end.  nn, of course,
%   is fully implemented.
% 
% * Seen articles are not implemented.  All articles are either read or
%   unread, available on your server or unavailable.  By default you will
%   only be able to read available unread articles.  This means preferences
%   relating to seen articles are not available.
% 
% * Parsing nn's kill/select file is unimplemented.  If you wish for certain
%   articles to be marked interesting, marked read, left alone or killed,
%   you will need to make a slrn scorefile and score appropriately.  All
%   related preferences are unavailable.
%   
% * Folder maintenance is not implemented.  This means slrn with this nn mode
%   is not able to browse through your folders deleting, saving, following up
%   to, replying to, etc. articles sitting on readable volumes.
% 
% * None of nn's command line options (-a0, -g, -r, -lN, etc.) are implemented.
% 
% * Nn's command mode command-line expansion is unimplemented.
% 
% * There is no prohibition on binding a key to a function.  nn doesn't allow
%   some functions to be bound to keys.
%
% * nn's colon-only commands (:admin, :bug, :cd, :compile, :coredump, etc.)
%   are unimplemented with the exception of :help (and that does not take
%   an argument yet), :toggle variable, :unset variable, :set variable value
%   and :x.
%
% * Some items remain unimplemented for now (opening and closing consolidated
%   menus) but will be implemented when the bugs are worked out of the basic
%   stuff.
%   
% * The screen will look different.  'slrn' will still be prominently
%   displayed and many slrn options will still work as expected.
%
% There are some other differences you probably won't care about in daily use:
% 
% * slrn uses real threads, nn uses subject sorting.  When the subject
%   of the articles in the thread doesn't change, you won't notice much
%   difference between the two.  If you have no idea what threads or
%   subject sorting means, don't worry about it.
%
% * nn makes the cursor-up & down keys wrap around the current menu of
%   articles.  This nn mode for slrn makes the cursor-up & down keys wrap
%   around the entire set of menus instead.
%   
% I am not yet committed to implementing or not implementing any particular
% feature, I take them on one by one as I read about them in the nn manpage
% and see them in action in an old installation of nn I have access to.  As
% slrn becomes more capable, I expect this macro set to evolve.  Eventually
% it will support many of nn's features (though not all).
%}}}

implements ("nn");

private variable t = 0;                          % a temporary variable

static variable
  slrnrc_variables = Assoc_Type[],
  prefs = Assoc_Type[],
  check_prefs = Assoc_Type[Ref_Type],
  commands = Assoc_Type[Ref_Type],

  % Exclusive modes--these need to be reset to 0 by all functions and checked
  % for by the header_number_hook().  It may be handy to make a function to
  % reset these variables to 0 then call that resetter function from most
  % functions.
  mode = Assoc_Type[Integer_Type];

% Read the selected articles then return to selection mode.
mode["read return"] = 0;
mode["open close subject"] = 0;
mode["is in range"] = 0;
mode["cancel"] = 0;

prefs["unsubscribe-mark-read"] = 1;

prefs["layout"] = 1;
private define check_layout (n)
{
   return (orelse { n < 0 } { n > 3 });
}
check_prefs["layout"] = &check_layout;
set_header_display_format (0, "%16r %60s +%l");
set_header_display_format (1, "%16r %l %60s");
set_header_display_format (2, "%6l %75s");
set_header_display_format (3, "%s");
for (t = 4; t < 10; t++) set_header_display_format (t, "");

% So the slrnrc settings nn mode sets can be reset back to the values the user
% had prior to loading this nn mode, use these functions instead of the Global
% alternatives.  See nn->reset() for more.
define set_integer_variable (v, d)                              %{{{
{
   !if (assoc_key_exists (slrnrc_variables, v))
     {
        slrnrc_variables[v] = get_variable_value (v);
     }
   Global->set_integer_variable (v, d);
}
%}}}
define set_string_variable (v, d)                               %{{{
{
   !if (assoc_key_exists (slrnrc_variables, v))
     {
        slrnrc_variables[v] = get_variable_value (v);
     }
   Global->set_string_variable (v, d);
}
%}}}

define version () { return (0.1); }
define set_preference (preference, value)                       %{{{
{
   !if (assoc_key_exists (prefs, preference))
     verror ("Preference does not exist: %S", preference);

   variable desired_type = typeof (prefs[preference]);
   if (typeof (value) != desired_type)
     verror ("Wrong type for %S: This preference wants %S not %S",
             preference, desired_type, typeof (value));

   if (@check_prefs[preference] (value))
     verror ("Bad setting for %S: %S\nSee documentation for information.",
             preference, value);

   prefs[preference] = value;
}
%}}}

%
% For internal use only.
% 

define is_selection_mode ()
{
   return (andelse
             { is_group_mode () == 0 }                % in article mode
             { (_is_article_visible () & 1) == 0 });  % no article visible
}

define is_reading_mode ()
{
   return (andelse
             { is_group_mode () == 0 }                % in article mode
             { _is_article_visible () & 1 }           % article is visible
             { is_article_window_zoomed () });
}

define enter_reading_mode ()
{
   % Show the current article fully zoomed.

   switch (_is_article_visible ())
     { case (0):
        call ("hide_article");
        !if (is_article_window_zoomed ()) call ("zoom_article_window");
     }
     { case (1):
        loop (2) call ("hide_article");
        !if (is_article_window_zoomed ()) call ("zoom_article_window");
     }
     { case (2):
        call ("hide_article");
        !if (is_article_window_zoomed ()) call ("zoom_article_window");
     }
     { case (3):
        !if (is_article_window_zoomed ()) call ("zoom_article_window");
     }
}

define reset_exclusive_modes ()
{
   variable k = NULL;
   foreach (mode) using ("key")
     {
        k = ();
        mode[k] = 0;
     }
}

%
% Exported functions not bound to a key are here.
% 
define unload_nn ()                                                  %{{{
{
   % Reset slrnrc settings back to what they were before nn mode was loaded.
   foreach (slrnrc_variables) using ("key", "value")
     {
        (var, val) = ();
        switch (typeof (val))
          { case (Integer_Type): Global->set_integer_variable (var, val); }
          { case (String_Type):  Global->set_string_variable (var, val); }
          {                      verror ("nn->%s: Unknown value type (%S).",
                                         _function_name (), val);
          }
     }

   % Reset exclusive mode settings for next invocation.
   reset_exclusive_modes ();
}
%}}}

%
% NN keybindable commands start here.
% 

define nn_continue ()
  % We can't name this function "continue" because that would conflict
  % with a function already defined in S-Lang.
{
   if (is_selection_mode ())                % next page|start reading mode
     {
        ERROR_BLOCK
          {
             _clear_error ();

             % Show first tagged article or go to next unread group.
             % If no next unread group, quit.
             call ("header_bob");
             if (orelse
                   { get_header_flags () & HEADER_TAGGED }
                   { next_tagged_header () })
               enter_reading_mode ();
             else
               {
                  call ("skip_to_next_group");
                  if (is_group_mode ())
                    {
                       set_input_chars ("y");
                       call ("quit");
                    }
               }
          }
        call ("pagedn");
     }
   else if (is_reading_mode ())
     % Show next page of selected article(s) or next unread group.
     % Return to same group if mode["read return"] is non-zero.
     {
        switch (get_next_art_pgdn_action ())
          { case (0):                            % next page of article
             call ("scroll_dn");
          }

          { case (1):                            % next unread article
             if (next_tagged_header ())
               {
                  call ("scroll_dn");
               }
             else
               {
                  if (mode["read return"])
                    {
                       call ("hide_article");
                       mode["read return"] = 0;
                    }
                  else
                    {
                       call ("skip_to_next_group");
                    }
               }
          }

          { case (2):                            % next group
             if (mode["read return"])
               {
                  call ("hide_article");
                  mode["read return"] = 0;
               }
             else
               {
                  call ("skip_to_next_group");
               }
          }
     }
}
definekey ("nn->nn_continue", " ", "article");
commands["continue"] = &nn->nn_continue;

define read_return ()
{
   if (is_selection_mode ())      % Show first tagged article or do nothing.
     {
        call ("header_bob");
        if (orelse
              { get_header_flags () & HEADER_TAGGED }
              { next_tagged_header () })
          {
             call ("scroll_dn");
             mode["read return"] = 1;
          }
     }
}
definekey ("nn->read_return", "Z", "article");
commands["read-return"] = &nn->read_return;

define read_skip ()
{
   if (is_selection_mode ())
     {
        call ("header_bob");
        if (orelse
              { get_header_flags () & HEADER_TAGGED }
              { next_tagged_header () })
          {
             enter_reading_mode ();
          }
        else
          {
             call ("skip_to_next_group");
          }
     }
   else if (is_reading_mode ())
     % The nn manpage says to "[m]ark all articles in the current group as
     % read and go directly to the next group" so I assume that means the
     % selected articles too.
     {
        if (get_yes_no_cancel ("Are you sure") == 1)
          {
             set_prefix_argument (1);            % remove all * tags
             call ("toggle_header_tag");

             call ("catchup");
             call ("skip_to_next_group");
          }
     }
}
definekey ("nn->read_skip", "X", "article");
commands["read-skip"] = &nn->read_skip;

define next_group ()
{
   if (orelse { is_selection_mode () } { is_reading_mode () })
     {
        call ("skip_to_next_group");
     }
}
definekey ("nn->next_group", "N", "article");
commands["next-group"] = &nn->next_group;

define select ()
{
   if (is_selection_mode ())
     {
        % This selects collapsed (or in nn parlance, consolidated) "threads".

        if (header_down (1))
          {
             () = header_up (1);
             call ("toggle_header_tag");
          }
        else
          {
             call ("toggle_header_tag");
             
             % nn really moves the cursor to the top of the current screen.
             % Since we can't do that, we'll settle for doing this.
             %
             % After playing around with this for a while, I'm not sure this
             % is a good idea.  I think it would make an acceptable thing for
             % a preference (default=off).
             
             
             call ("header_bob");
          }
     }
}
definekey ("nn->select", ".", "article");
commands["select"] = &nn->select;

define header_number_hook ()
{
   if (is_selection_mode ())
     {
        select ();
     }
}

define line_minus_1 ()
{
   if (is_selection_mode ()) call ("up");
}
definekey ("nn->line_minus_1", "/", "article");
definekey ("nn->line_minus_1", "\eOA", "article");
definekey ("nn->line_minus_1", "\e[A", "article");
commands["line-1"] = &line_minus_1;

define select_invert ()
{
   if (is_selection_mode ())
     {
        variable
          starting_place = extract_article_header ("Message-ID"),
          msgid = "";

        call ("header_bob");
        do
          {
             msgid = extract_article_header ("Message-ID");
             call ("toggle_header_tag");
          }
        while (extract_article_header ("Message-ID") != msgid);

        () = locate_header_by_msgid (starting_place, 0);
     }
}
definekey ("nn->select_invert", "@", "article");
commands["select-invert"] = &nn->select_invert;

define unselect_all ()
{
   if (is_selection_mode ())
     {
        set_prefix_argument (1);
        call ("toggle_header_tag");
     }
}
definekey ("nn->unselect_all", "~", "article");
commands["unselect-all"] = &nn->unselect_all;

define page_plus_1 ()
{
   if (is_selection_mode ())
     {
        ERROR_BLOCK
          {
             _clear_error ();
             call ("header_bob");
          }
        call ("pagedn");
     }
}
definekey ("nn->page_plus_1", ">", "article");
commands["page+1"] = &nn->page_plus_1;

define page_minus_1 (for_reading_mode_only)
{
   if (is_selection_mode ())
     {
        if (for_reading_mode_only) return ();

        ERROR_BLOCK
          {
             _clear_error ();
             call ("header_eob");
          }
        call ("pageup");
     }
   else if (is_reading_mode ())
     {
        !if (for_reading_mode_only) return ();
        call ("scroll_up");
     }
}
definekey ("nn->page_minus_1(0)", "<", "article");    % for selection mode only
definekey ("nn->page_minus_1(1)", "^?", "article");   % for reading mode only
definekey ("nn->page_minus_1(1)", "^H", "article");   % for reading mode only
commands["page-minus-1"] = &nn->page_minus_1;

define goto_first_menu_page ()
{
   if (is_selection_mode ()) call ("header_bob");
}
definekey ("nn->goto_first_menu_page", "^", "article");
commands["goto-first-menu-page"] = &nn->goto_first_menu_page;

define goto_last_menu_page ()
{
   if (is_selection_mode ()) call ("header_eob");
}
definekey ("nn->goto_last_menu_page", "$", "article");
commands["goto-last-menu-page"] = &nn->goto_last_menu_page;

define page_minus_1 ()
{
   if (is_reading_mode ())
     {
        call ("scroll_up");
     }
}
definekey ("nn->page_minus_1", "^?", "article");
definekey ("nn->page_minus_1", "^H", "article");
commands["page-1"] = &nn->page_minus_1;

define line_plus_1 ()
{
   if (is_reading_mode ())
     {
        call ("article_linedn");
     }
   else if (is_selection_mode ())
     {
        call ("down");
     }
}
definekey ("nn->line_plus_1", ",", "article");
definekey ("nn->line_plus_1", "\e[B", "article");
definekey ("nn->line_plus_1", "\eOB", "article");
definekey ("nn->line_plus_1", "\r", "article");
commands["line+1"] = &nn->line_plus_1;

define skip_lines ()
{
   if (is_reading_mode ())
     {
        % This isn't exactly what nn does, but it's close enough for now.
        call ("skip_quotes");
     }
}
definekey ("nn->skip_lines", "\t", "article");
commands["skip-lines"] = &nn->skip_lines;

define page_equals_0 ()
{
   if (is_reading_mode ())
     {
        % slrn's toggle_headers has a side effect of jumping to the top of
        % article when *showing* all headers (not when hiding full headers
        % to show visible_headers).  So we don't need an explicit art_bob
        % here.
        if (headers_hidden_mode ()) call ("toggle_headers");
     }
}
definekey ("nn->page_equals_0", "h", "article");
commands["page=0"] = &nn->page_equals_0;

define rot13 ()
{
   if (is_reading_mode ())
     {
        call ("toggle_rot13");
     }
}
definekey ("nn->rot13", "D", "article");
commands["rot13"] = &nn->rot13;

define next_article ()
{
   if (is_reading_mode ())
     {
        if (next_tagged_header ()) call ("scroll_dn");
        else call ("skip_to_next_group");
     }
}
definekey ("nn->next_article", "n", "article");
commands["next-article"] = &nn->next_article;

define previous ()
{
   if (is_reading_mode ())
     {
        if (prev_tagged_header ()) call ("scroll_dn");
     }
}
definekey ("nn->previous", "p", "article");
commands["previous"] = &nn->previous;

define advance_article ()
{
   if (is_reading_mode ())
     {
        if (is_thread_collapsed ())
          {
             uncollapse_thread ();
          }
        if (header_down (1)) call ("scroll_dn");
     }
}
definekey ("nn->advance_article", "a", "article");
commands["advance-article"] = &nn->advance_article;

define back_article ()
{
   if (is_reading_mode ())
     {
        !if (header_up (1)) return ();
        if (is_thread_collapsed ())
          {
             uncollapse_thread ();
             () = header_down (thread_size () - 1);
          }
        call ("scroll_dn");
     }
}
definekey ("nn->back_article", "b", "article");
commands["back-article"] = &nn->back_article;

define goto_menu ()
{
   if (is_reading_mode ())
     {
        call ("hide_article");
     }
}
definekey ("nn->goto_menu", "=", "article");
commands["goto-menu"] = &nn->goto_menu;

define reply ()
{
   if (orelse { is_selection_mode () } { is_reading_mode () }) call ("reply");
}
definekey ("nn->reply", "R", "article");
commands["reply"] = &nn->reply;

define follow ()
{
   if (orelse { is_selection_mode () } { is_reading_mode () }) call ("followup");
}
definekey ("nn->follow", "F", "article");
commands["follow"] = &nn->follow;

define post ()
{
   if (orelse { is_selection_mode () } { is_reading_mode () }) call ("post");
}
commands["post"] = &nn->post;

define colon_command ()
{
   variable cmd = read_mini ("", "", "");        % ": " is the prompt.
   !if (strlen (cmd)) return ();
   if (assoc_key_exists (cmd, commands)) eval (@commands[cmd]);
   else verror ("unknown command: \"%s\"", cmd);
}
definekey ("nn->colon_command", ":", "article");

define unsub ()
{
   % Set slrn's group mode cursor to the current newsgroup and unsubscribe.
   if (orelse { is_reading_mode () } { is_selection_mode () })
     {
        set_group_flags (get_group_flags () | GROUP_UNSUBSCRIBED);
        if (prefs["unsubscribe-mark-read"]) call ("catchup");
     }
}
definekey ("nn->unsub", "U", "article");
commands["unsub"] = &nn->unsub;

define cancel ()
{
   % CANCEL_MODE is only used to tell the header_number_hook() not to do
   % anything with the article (such as toggle selection or enter read mode).
   % This lets us position the cursor on the article and call("cancel").

   if (is_selection_mode ())
     {
        message_now ("Enter identifier of article to cancel.");
        mode["cancel"] = 1;
        call ("cancel");
     }
   else if (is_reading_mode ())
     {
        call ("cancel");
     }
}
definekey ("nn->cancel", "C", "article");
commands["cancel"] = &nn->cancel;

define overview ()
{
   % Show groups with unread articles.
   variable starting_group = current_newsgroup ();
   if (is_article_mode ()) call ("quit");
   !if (strlen (starting_newsgroup)) return ();

   % Hide unsubscribed groups.
   set_prefix_argument (1);
   call ("toggle_list_all");
   reset_prefix_arg ();

   % Hide 0-article groups.
   call ("bob");
   do
     {
        !if (group_unread ())
          {
             call ("toggle_hidden");
             !if (group_unread ()) call ("toggle_hidden");
             break;
          }
     }
   while (group_down_n (1));
   !if (Enhanced->simple_group_search (starting_group))
     error ("nn: Unexpected inability to find " + starting_group);
}
definekey ("nn->overview", "Y", "article");
commands["overview"] = &nn->overview;

define layout ()
{
   call ("toggle_header_formats");
}
definekey ("nn->layout", "\"", "article");
commands["layout"] = &nn->layout;

define help ()
{
   call ("help");
}
commands["help"] = &nn->help;

define toggle (v)
{
   % To do: maintain the exception list for non-toggled integer preference

   variable non_toggled_preferences = Assoc_Type[Null_Type];
   non_toggled_preferences["layout"] = NULL;     % I don't like this approach.

   if (typeof (v) != String_Type)
     {
        verror ("Preference must be a string: %S", v);
     }
   !if (assoc_key_exists (prefs, v))
     {
        verror ("No such variable %S", v);
     }
   if (typeof (prefs[v]) != Integer_Type)
     {
        verror ("Variable %S is not toggleable.", v);
     }
   if (assoc_key_exists (non_toggled_preferences, v))
     {
        verror ("Variable %S is not a toggle.", v);
     }

   if (prefs[v]) { set_preference (v, 0); }
   else { set_preference (v, 1); }
}
commands["toggle"] = &nn->toggle;

% To be used in the future (move to top of file when used)

variable slrnrc_variables = Assoc_Type[];

% So the slrnrc settings nn mode sets can be reset back to the values the user
% had prior to loading this nn mode, use these functions instead of the Global
% alternatives.  See nn->reset() for more.
private define set_integer_variable (v, d)                              %{{{
{
   !if (assoc_key_exists (slrnrc_variables, v))
     {
        slrnrc_variables[v] = get_variable_value (v);
     }
   Global->set_integer_variable (v, d);
}
%}}}
private define set_string_variable (v, d)                               %{{{
{
   !if (assoc_key_exists (slrnrc_variables, v))
     {
        slrnrc_variables[v] = get_variable_value (v);
     }
   Global->set_string_variable (v, d);
}
%}}}
